/*-*- Mode: C; c-basic-offset: 8; indent-tabs-mode: nil -*-*/
#include "config.h"
#include <syslog.h>
#include <sys/time.h>
#include <time.h>
#include <stdio.h>
#include <sys/stat.h>
#include "log.h"
#include "button.h"
#include "led.h"
#include "touch_sx9512.h"
#include "prox_px3220.h"

#define BUTTON_STATE_MAX  3

typedef enum {
    BUTTON_NORMAL,
    BUTTON_PRODUCTION,
    BUTTON_CUSTOM, /*Not Implemented*/
} button_mode_t;


static time_t starttime_t;
static int global_button_state = BUTTON_NORMAL;

static struct ubus_context *global_ubus_ctx;
static struct blob_buf bblob;

static char *config_file;       /* path+name of the buttons config file */
static time_t config_mtime_sec; /* mtime of config file */

static void reread_config( void );

/* Names for button_state */
static const char* const buttons_mode[BUTTON_STATE_MAX] =
{ "normal", "production" , "custom"};

/* we find out the index for a match in an array of char pointers containing max number of pointers */
static int get_index_by_name(const char *const*array, int max, const char *name)
{
	int i;
	for (i=0; i < max ; i++ ){
		if (!strcasecmp(name, array[i]))
			return i;
	}
	return -1;
}

static void button_ubus_interface_event(struct ubus_context *ubus_ctx, char *button, button_state_t pressed, bool longpress)
{
	char s[UBUS_BUTTON_NAME_PREPEND_LEN+BUTTON_MAX_NAME_LEN+1];
	s[0]=0;
	strcat(s, UBUS_BUTTON_NAME_PREPEND);
	strcat(s, button);
	memset(&bblob, 0, sizeof(bblob));
	blob_buf_init(&bblob, 0);
	blobmsg_add_string(&bblob, "action", pressed ? "pressed" : "released");
	blobmsg_add_string(&bblob, "type", longpress ? "long" : "short");
	ubus_send_event(ubus_ctx, s, bblob.head);
        blob_buf_free(&bblob);
}


/* used to map in the driver buttons to a function button */
struct button_drv_list {
        struct list_head list;
        struct timespec pressed_time;
        struct button_drv *drv;
};

/**/
struct function_button {
	struct list_head list;
        char *name;
        int dimming;
        char *hotplug;
	char *hotplug_long;
        int enable;                             /* should this button be sent to the system or not */
        int fixed;                              /* if set prevents the config from beeing changed dynamically */
        int minpress;
        int longpress;                          /* negative value means valid if  mintime < time < abs(longpress ) */
                                                /* positive value means valid if time > longpreass */
                                                /* zero value means valid if time > mintime */
	struct list_head drv_list;              /* list of all driver button that is needed to activate this button function */
};

/* PUT every button from drivers into a list */
struct drv_button_list{
	struct list_head list;
	struct button_drv *drv;
};

/* list of all driver buttons added by drivers. */
static LIST_HEAD(drv_buttons_list);

/* list containing all function buttons read from config file */
static LIST_HEAD(buttons);

enum {
	BUTTON_MODE,
	__BUTTON_MAX
};

static const struct blobmsg_policy button_policy[] = {
	[BUTTON_MODE]      = { .name = "mode",		.type = BLOBMSG_TYPE_STRING },
};


void button_add( struct button_drv *drv)
{
	struct drv_button_list *drv_node = malloc(sizeof(struct drv_button_list));

	DBG(1,"called with button name [%s]", drv->name);
	drv_node->drv = drv;

	list_add(&drv_node->list, &drv_buttons_list);
}

static struct button_drv *get_drv_button(const char *name)
{
	struct list_head *i;
	list_for_each(i, &drv_buttons_list) {
		struct drv_button_list *node = list_entry(i, struct drv_button_list, list);
                if (! strcmp(node->drv->name, name))
                        return node->drv;
	}
        return NULL;
}

static struct function_button *get_button(const char *name)
{
	struct list_head *i;
	list_for_each(i, &buttons) {
		struct function_button *node = list_entry(i, struct function_button, list);
                if (! strcmp(node->name, name))
                        return node;
	}
        return NULL;
}

//! Read state for single button
static button_state_t read_button_state(const char *name)
{
	struct list_head *i;

	/* sx9512 driver needs to read out all buttons at once */
	/* so call it once at beginning of scanning inputs  */
	sx9512_check();

#ifdef BRCM_BOARD
	/* same for px3220 */
	px3220_check();
#endif
	list_for_each(i, &buttons) {
		struct list_head *j;
		struct function_button *node = list_entry(i, struct function_button, list);
		if(!strcmp(node->name, name)) {
			button_state_t state=BUTTON_ERROR;
			list_for_each(j, &node->drv_list) {
				struct button_drv_list *drv_node = list_entry(j, struct button_drv_list, list);
				if(drv_node->drv) {
					if(drv_node->drv->func->get_state(drv_node->drv))
						return BUTTON_PRESSED;
					else
						state=BUTTON_RELEASED;
				}
			}
			return state;
		}
	}
	return BUTTON_ERROR;
}

struct button_status {
	char name[BUTTON_MAX_NAME_LEN];
	button_state_t state;
};

struct button_status_all {
	int n;
	struct button_status status[BUTTON_MAX];
};


//! Read states for all buttons
static struct button_status_all * read_button_states(void)
{
	static struct button_status_all p;
	struct list_head *i;
        p.n=0;
	/* sx9512 driver needs to read out all buttons at once */
	/* so call it once at beginning of scanning inputs  */
	sx9512_check();
#ifdef BRCM_BOARD
	/* same for px3220 */
	px3220_check();
#endif
	list_for_each(i, &buttons) {
		struct list_head *j;
		button_state_t state=BUTTON_ERROR;
		struct function_button *node = list_entry(i, struct function_button, list);
		strcpy(p.status[p.n].name, node->name);
		list_for_each(j, &node->drv_list) {
			struct button_drv_list *drv_node = list_entry(j, struct button_drv_list, list);
			if(drv_node->drv) {
				if(drv_node->drv->func->get_state(drv_node->drv))
					state=BUTTON_PRESSED;
				else
					state=BUTTON_RELEASED;
			}
		}
		p.status[p.n].state = state;
		p.n++;
	}
	return &p;
}

static void dump_drv_list(void)
{
	struct list_head *i;
	list_for_each(i, &drv_buttons_list) {
		struct drv_button_list *node = list_entry(i, struct drv_button_list, list);
		DBG(1,"button name = [%s]",node->drv->name);
	}
}

static void dump_buttons_list(void)
{
	struct list_head *i;
	list_for_each(i, &buttons) {
		struct function_button *node = list_entry(i, struct function_button, list);
		DBG(1,"button name = [%s]",node->name);
                {
                        struct list_head *j;
                        list_for_each(j, &node->drv_list) {
                                struct button_drv_list *drv_node = list_entry(j, struct button_drv_list, list);
                                if(drv_node->drv != NULL)
                                        DBG(1,"%13s drv button name = [%s]","",drv_node->drv->name);
                        }
                        DBG(1,"%13s minpress = %d","",node->minpress);
                        DBG(1,"%13s longpress = %d","",node->longpress);
                }
	}
}


//! Run the hotplug command associated with function button
//! @retval 0 ok
static int button_hotplug_cmd(const char *name, bool longpress)
{
	struct list_head *i;
	list_for_each(i, &buttons) {
		struct function_button *node = list_entry(i, struct function_button, list);
		if(!strcmp(node->name, name)) {
			char str[512];
			char *hotplug = node->hotplug;
			if(longpress && node->hotplug_long)
				hotplug = node->hotplug_long;
			if(!hotplug)
				return 1;
			DBG(1, "send key %s [%s] to system %s", node->name, hotplug, longpress ? "(longpress)" : "");
			snprintf(str, 512, "ACTION=register INTERFACE=%s /sbin/hotplug-call button &", hotplug);
			system(str);
			syslog(LOG_INFO, "%s",str);
			return 0;
		}
	}
	DBG(1,"unknown button: \"%s\"", name);
	return 1;
}


static int timer_started(struct button_drv_list *button_drv)
{
        if (button_drv->pressed_time.tv_sec == 0 )
                if (button_drv->pressed_time.tv_nsec == 0 )
                        return 0;
        return 1;
}

static void timer_start(struct button_drv_list *button_drv)
{
        clock_gettime(CLOCK_MONOTONIC, &button_drv->pressed_time);
}

static void timer_stop(struct button_drv_list *button_drv)
{
        button_drv->pressed_time.tv_sec = 0;
        button_drv->pressed_time.tv_nsec = 0;
}


static button_press_type_t timer_valid(struct button_drv_list *button_drv, int mtimeout, int longpress)
{
        struct timespec now;
        int sec;
        int nsec;
	int time_elapsed;

        if (timer_started(button_drv)) {
                clock_gettime(CLOCK_MONOTONIC, &now);
                sec =  now.tv_sec  - button_drv->pressed_time.tv_sec;
                nsec = now.tv_nsec - button_drv->pressed_time.tv_nsec;
		time_elapsed = sec*1000 + nsec/1000000;
                if ( mtimeout < time_elapsed) {
			if (longpress && (longpress < time_elapsed))
				return BUTTON_PRESS_LONG;
			return BUTTON_PRESS_SHORT;
                }
        }
	return BUTTON_PRESS_NONE;
}

#define BUTTON_TIMEOUT 100
static void button_handler(struct uloop_timeout *timeout);
static struct uloop_timeout button_inform_timer = { .cb = button_handler };

static void button_handler(struct uloop_timeout *timeout)
{
 	struct list_head *i;
	int r;
        struct stat stat_buf;
//        DBG(1, "");

        /* sx9512 driver needs to read out all buttons at once */
        /* so call it once at beginning of scanning inputs  */
        sx9512_check();

#ifdef BRCM_BOARD
        /* same for px3220 */
        px3220_check();
#endif

        /* check to see if the config files has been changed */
        if (0 == stat(config_file, &stat_buf)) {
                if (config_mtime_sec != stat_buf.st_mtim.tv_sec) {
                        reread_config();
                        config_mtime_sec = stat_buf.st_mtim.tv_sec;
                }
        }

        /* clean out indicator status, set by any valid press again if we find it */
        led_pressindicator_set(PRESS_NONE);

	list_for_each(i, &buttons) {
                struct list_head *j;
		struct function_button *node = list_entry(i, struct function_button, list);

                list_for_each(j, &node->drv_list) {
                        struct button_drv_list *drv_node = list_entry(j, struct button_drv_list, list);
                        if (drv_node->drv) {
                                button_press_type_t time_type;
                                button_state_t st = drv_node->drv->func->get_state(drv_node->drv);

                                if (st == BUTTON_PRESSED ) {
                                        if (! timer_started(drv_node)) {
                                                timer_start(drv_node);
                                                DBG(1, " %s pressed", drv_node->drv->name);
//						button_ubus_interface_event(global_ubus_ctx, node->name, BUTTON_PRESSED);
                                        }

                                        time_type = timer_valid(drv_node, node->minpress, node->longpress);
                                        if (node->enable) {
                                                if( time_type == BUTTON_PRESS_LONG )
                                                        led_pressindicator_set(PRESS_LONG);
                                                if( time_type == BUTTON_PRESS_SHORT )
                                                        led_pressindicator_set(PRESS_SHORT);
                                        }
                                }

                                if (st == BUTTON_RELEASED ) {
                                        if (timer_started(drv_node)) {
                                                DBG(1, " %s released", drv_node->drv->name);
                                                if((r=timer_valid(drv_node, node->minpress, node->longpress))) {
                                                        DBG(1, " %s released timer_valid=%d", drv_node->drv->name,r);

							if(node->dimming)
								led_dimming();

                                                        if (node->enable) {
                                                                struct timeval tv;
                                                                /* filter out button presses that are triggered less the 4 seconds since start */
                                                                gettimeofday(&tv, NULL);
                                                                if (tv.tv_sec > starttime_t + 4) {
									if(global_button_state == BUTTON_NORMAL) {  /* Enter only in normal mode */
										   DBG(1,"button_handler function is in Normal mode");
		                                                                   button_ubus_interface_event(global_ubus_ctx,
				                                                                               node->name,
		                                                                                               BUTTON_RELEASED,
                                                                                                               r==BUTTON_PRESS_LONG);
                                                                                   button_hotplug_cmd(node->name, r==BUTTON_PRESS_LONG);
                                                                        }else
									     syslog(LOG_INFO, "In Production Mode: Button event from %s", node->name);
								}
                                                                else
                                                                   syslog(LOG_WARNING, "button %s press ignored", drv_node->drv->name);
                                                        }
						}
                                        }
                                        timer_stop(drv_node);
                                }

                                if (st == BUTTON_ERROR ) {
                                        timer_stop(drv_node);
                                        DBG(3, " %s state = BUTTON_ERROR", drv_node->drv->name);
                                }
                        }
                }
        }
	uloop_timeout_set(&button_inform_timer, BUTTON_TIMEOUT);
}


static int button_state_method(struct ubus_context *ubus_ctx, struct ubus_object *obj,
	struct ubus_request_data *req, const char *method, struct blob_attr *msg)
{
//	button_state_t state = read_button_state(obj->name+UBUS_BUTTON_NAME_PREPEND_LEN);
	memset(&bblob, 0, sizeof(bblob));
	blob_buf_init(&bblob, 0);
	switch(read_button_state(obj->name+UBUS_BUTTON_NAME_PREPEND_LEN)) {
	case BUTTON_RELEASED:
		blobmsg_add_string(&bblob, "state", "released");
		break;
	case BUTTON_PRESSED:
		blobmsg_add_string(&bblob, "state", "pressed");
		break;
	default:
		blobmsg_add_string(&bblob, "state", "error");
	}
	ubus_send_reply(ubus_ctx, req, bblob.head);
        blob_buf_free(&bblob);
	return 0;
}


static int button_press_method(struct ubus_context *ubus_ctx, struct ubus_object *obj,
				struct ubus_request_data *req, const char *method, struct blob_attr *msg)
{
	memset(&bblob, 0, sizeof(bblob));
	blob_buf_init(&bblob, 0);

        syslog(LOG_WARNING, "button_press_method on %s", obj->name);

	if(button_hotplug_cmd(obj->name+UBUS_BUTTON_NAME_PREPEND_LEN, 0))
		blobmsg_add_string(&bblob, "return", "error");
	else
		blobmsg_add_string(&bblob, "return", "ok");
	ubus_send_reply(ubus_ctx, req, bblob.head);
        blob_buf_free(&bblob);
	return 0;
}


static int button_press_long_method(struct ubus_context *ubus_ctx, struct ubus_object *obj,
			       struct ubus_request_data *req, const char *method, struct blob_attr *msg)
{
	memset(&bblob, 0, sizeof(bblob));
	blob_buf_init(&bblob, 0);
	if(button_hotplug_cmd(obj->name+UBUS_BUTTON_NAME_PREPEND_LEN, 1))
		blobmsg_add_string(&bblob, "return", "error");
	else
		blobmsg_add_string(&bblob, "return", "ok");
	ubus_send_reply(ubus_ctx, req, bblob.head);
        blob_buf_free(&bblob);
	return 0;
}


static int buttons_state_method(struct ubus_context *ubus_ctx, struct ubus_object *obj,
	struct ubus_request_data *req, const char *method, struct blob_attr *msg)
{
	int i;
	static struct button_status_all *p;
	p = read_button_states();
	memset(&bblob, 0, sizeof(bblob));
	blob_buf_init(&bblob, 0);
	for(i=0;i < p->n; i++) {
		switch(p->status[i].state) {
		case BUTTON_RELEASED:
			blobmsg_add_string(&bblob, p->status[i].name, "released");
			break;
		case BUTTON_PRESSED:
			blobmsg_add_string(&bblob, p->status[i].name, "pressed");
			break;
		default:
			blobmsg_add_string(&bblob, p->status[i].name, "error");
		}
	}
	ubus_send_reply(ubus_ctx, req, bblob.head);
        blob_buf_free(&bblob);
	return 0;
}

static int buttons_mode_check(struct ubus_context *ubus_ctx, struct ubus_object *obj,
	struct ubus_request_data *req, const char *method, struct blob_attr *msg)
{
	memset(&bblob, 0, sizeof(bblob));

	blob_buf_init(&bblob, 0);

	switch(global_button_state) {
        case BUTTON_PRODUCTION:
		blobmsg_add_string(&bblob, "mode", "production");
		break;
	case BUTTON_NORMAL:
		blobmsg_add_string(&bblob, "mode", "normal");
		break;
	case BUTTON_CUSTOM:
		blobmsg_add_string(&bblob, "mode", "custom");
		break;
	default:
		blobmsg_add_string(&bblob, "mode", "invalid-mode");
	}
	ubus_send_reply(ubus_ctx, req, bblob.head);
        blob_buf_free(&bblob);
	return 0;
}

static int buttons_set_method(struct ubus_context *ubus_ctx, struct ubus_object *obj,
                          struct ubus_request_data *req, const char *method,
                          struct blob_attr *msg)
{

	struct blob_attr *tb[__BUTTON_MAX];

	blobmsg_parse(button_policy, ARRAY_SIZE(button_policy), tb, blob_data(msg), blob_len(msg));

	if (tb[BUTTON_MODE]) {

		char* mode;
		int state_idx;

		mode = blobmsg_data(tb[BUTTON_MODE]);

		state_idx = get_index_by_name(buttons_mode, BUTTON_STATE_MAX , mode);
		if (state_idx == -1) {
			syslog(LOG_WARNING, "button_set_method: Unknown state %s.", mode);
			return 0;
		}

		if (!strcmp(mode,"normal") ) {
			global_button_state = BUTTON_NORMAL;
		}

		if (!strcmp(mode,"production") ) {
			global_button_state = BUTTON_PRODUCTION;
		}

		DBG(1,"button global state set to [%s]", buttons_mode[state_idx]);

	}else
		syslog(LOG_WARNING, "Unknown attribute [%s]", (char *)blob_data(msg));

	return 0;

}

static const struct ubus_method button_methods[] = {
//	{ .name = "status", .handler = button_status_method },
	{ .name = "state", .handler = button_state_method },
	{ .name = "press", .handler = button_press_method },
	{ .name = "press_long", .handler = button_press_long_method },
};

static struct ubus_object_type button_object_type = UBUS_OBJECT_TYPE("button", button_methods);


static const struct ubus_method buttons_methods[] = {
	UBUS_METHOD("set",   buttons_set_method, button_policy),
	{ .name = "state", .handler = buttons_state_method },
	{ .name = "mode",   .handler = buttons_mode_check },
};

static struct ubus_object_type buttons_object_type = UBUS_OBJECT_TYPE("buttons", buttons_methods);

static struct ubus_object buttons_object = { .name = "buttons", .type = &buttons_object_type, .methods = buttons_methods, .n_methods = ARRAY_SIZE(buttons_methods), };


static void add_buttons_to_ubus(struct server_ctx *s_ctx)
{

	int ret = ubus_add_object(s_ctx->ubus_ctx, &buttons_object);
		if (ret)
			DBG(1,"Failed to add object: %s", ubus_strerror(ret));
       /* struct ubus_object *ubo;
        int r;
        char name[UBUS_BUTTON_NAME_PREPEND_LEN+BUTTON_MAX_NAME_LEN+1];

	struct list_head *i;
	list_for_each(i, &buttons) {
		struct function_button *node = list_entry(i, struct function_button, list);
                {
                        /* register buttons object with ubus */
                     /*   if((r = ubus_add_object(s_ctx->ubus_ctx, &buttons_object)))
                                DBG(1,"Failed to add object: %s", ubus_strerror(r));

                        ubo = malloc(sizeof(struct ubus_object));
                        memset(ubo, 0, sizeof(struct ubus_object));

                        snprintf(name, UBUS_BUTTON_NAME_PREPEND_LEN+BUTTON_MAX_NAME_LEN+1, "%s%s",
                                 UBUS_BUTTON_NAME_PREPEND,
                                 node->name);
                        ubo->name      = strdup(node->name);
                        ubo->methods   = button_methods;
                        ubo->n_methods = ARRAY_SIZE(button_methods);
                        ubo->type      = &button_object_type;
                        if((r = ubus_add_object(s_ctx->ubus_ctx, ubo)))
                                DBG(1,"Failed to add object: %s", ubus_strerror(r));
                }
	}*/
}

static void create_button(struct uci_context *ctx, struct uci_section *sec, int fixed)
{
        struct function_button *function;
        struct uci_option *opt;
	int default_minpress = 100;
        const char *s;

        if (! sec->e.name)
        {
                syslog(LOG_WARNING, "Button config without a name");
                return;
        }

        function = malloc(sizeof(struct function_button));
        memset(function,0,sizeof(struct function_button));

        function->name = strdup(sec->e.name);

        /* read out enable */
        s = uci_lookup_option_string(ctx, sec, "enable");
        DBG(1, "%s: enable = [%s]", function->name, s);
        if (s){
                function->enable =  strtol(s,0,0);
        }else
                function->enable =  1; /* if value do not exist its assumed to be enabled */

        /* read out dimming */
        s = uci_lookup_option_string(ctx, sec, "dimming");
        DBG(1, "%s: dimming = [%s]", function->name, s);
        if (s){
                function->dimming = 1;
        }else
                function->dimming = 0;

        /* read out minpress */
        s = uci_lookup_option_string(ctx, sec, "minpress");
        DBG(1, "%s: minpress = [%s]", function->name, s);
        if (s){
                function->minpress =  strtol(s,0,0);
        }else
                function->minpress =  default_minpress;

        /* read out long_press */
        s = uci_lookup_option_string(ctx, sec, "longpress");
        DBG(1, "%s: longpress = [%s]", function->name, s);
        if (s){
                function->longpress =  strtol(s,0,0);
        }

        /* read out hotplug option */
        s = uci_lookup_option_string(ctx, sec, "hotplug");
        DBG(1, "%s: hotplug = [%s]", function->name, s);
        if (s){
                function->hotplug = strdup(s);
        }

        /* read out hotplug option for longpress */
        s = uci_lookup_option_string(ctx, sec, "hotplug_long");
        DBG(1, "%s: hotplug_long = [%s]", function->name, s);
        if (s){
                function->hotplug_long = strdup(s);
        }

        /* read out list of driver buttons that need to be active to activate this
           button */
        opt = uci_lookup_option(ctx, sec, "button");
        DBG(1, "%s: button driver = option", function->name);
        if (opt) {
                struct uci_element *e;
                int num = 0;
                DBG(1, "option type %x ",(int)opt->type);
                DBG(1, "option section %p ",opt->section);
                DBG(1, "option elem-name[%s]",opt->e.name);

                INIT_LIST_HEAD(&function->drv_list);

                uci_foreach_element(&opt->v.list, e) {
                        struct button_drv_list *new_button;
                        DBG(1,"e type %x",e->type);
                        DBG(1,"e name %s",e->name);

                        new_button = malloc(sizeof(struct button_drv_list));
                        memset(new_button,0,sizeof(struct button_drv_list));
                        DBG(1," A %d %d",sizeof(struct button_drv_list),sizeof(new_button));

                        new_button->drv = get_drv_button(e->name);

                        if(new_button->drv == NULL) {
                                syslog(LOG_WARNING, "%s wanted drv button [%s] but it can't be found. check spelling.",
                                       function->name,
                                       e->name);
                                free(new_button);
                        }else {
                                num++;
                                list_add( &new_button->list, &function->drv_list);
                        }
                }

                if (num == 0)
                        syslog(LOG_WARNING, "Function  %s did not have any mapping to a driver button", function->name);

        }else
                syslog(LOG_WARNING, "Function  %s did not have any mapping to a driver button", function->name);

        if (fixed)
                function->fixed=1;

        /* if name not already in list add it. */
        if ( ! get_button(function->name) )
                list_add(&function->list, &buttons);

}

static void button_from_hw(struct server_ctx *s_ctx)
{
        int ret;
        struct uci_ptr uci_p;
	struct uci_element *elem;

	memset(&uci_p, 0, sizeof(uci_p));
	uci_p.package = "hw";

        ret = uci_lookup_ptr(s_ctx->uci_ctx, &uci_p, NULL, false);

        /* loop over any button section in /lib/db/config/hw file */
        if (ret == UCI_OK ){
                uci_foreach_element(&uci_p.p->sections, elem) {
                        struct uci_section *sec = uci_to_section(elem);
                        if ( ! strcmp(sec->type, "button")) {
                                DBG(1,"Found button named %s",sec->e.name);
                                create_button(s_ctx->uci_ctx, sec, 1);
                        }
                }
        }else{
                DBG(1, "Failed to find any button section in /lib/db/config/hw");
        }
}

static void button_from_config(struct server_ctx *s_ctx)
{
	struct uci_element *elem;
        struct stat stat_buf;
        char config_file_name[1024];
	struct uci_package *button_package = NULL;
        struct uci_context *ctx = uci_alloc_context();

        snprintf(config_file_name, 1024, "%s/%s", s_ctx->config_path, "buttons");
        config_file = strdup(config_file_name);

	if (uci_load(ctx, config_file, &button_package))
		return;

        /* save the mtime of the config file so we know later if it has changed */
        if (-1 == stat(config_file, &stat_buf)) {
                syslog(LOG_WARNING, "Could not stat button %s", config_file);
                return;
        }
        config_mtime_sec = stat_buf.st_mtim.tv_sec;

        /* loop over any button section in /etc/config/button */
        uci_foreach_element(&button_package->sections, elem) {
                struct uci_section *sec = uci_to_section(elem);
                if ( ! strcmp(sec->type, "button")) {
                        DBG(1,"Found button named %s",sec->e.name);
                        create_button(s_ctx->uci_ctx, sec, 0);
                }
        }
}


void button_init( struct server_ctx *s_ctx)
{

        DBG(1, "Read buttons from /lib/db/config/hw");
        button_from_hw(s_ctx);

        DBG(1, "Read buttons from /etc/config/buttons");
        button_from_config(s_ctx);

        dump_drv_list();
        dump_buttons_list();

	global_ubus_ctx=s_ctx->ubus_ctx;
        uloop_timeout_set(&button_inform_timer, BUTTON_TIMEOUT);

	add_buttons_to_ubus(s_ctx);

        return;
}

/* we are only interested in reading out the enable option for all buttons */
static void reread_config( void )
{
        struct uci_context *button_ctx = uci_alloc_context();
	struct uci_package *button_package = NULL;
	struct uci_element *elem;

        if (uci_load(button_ctx, config_file, &button_package)){
                uci_free_context(button_ctx);
                return;
        }
        uci_foreach_element(&button_package->sections, elem) {
                struct uci_section *sec = uci_to_section(elem);
                int new_enable = 0;
                struct function_button *button;
                const char * s;

                if (! sec->e.name)
                {
                        syslog(LOG_WARNING, "Button config without a name, check %s",
                               config_file);
                        continue;
                }

                /* read out enable */
                s = uci_lookup_option_string(button_ctx, sec, "enable");
                if (s){
                        new_enable =  strtol(s,0,0);
                }else
                        new_enable =  1; /* if value do not exist its assumed to be enabled */

                /* get current value and change it */
                button = get_button(sec->e.name);
                if (button) {
                        if (button->enable != new_enable && button->fixed == 0){
                                button->enable = new_enable;
                                DBG(1, "%s: change enable to = %d", button->name, new_enable);
                        }
                }
        }

        uci_free_context(button_ctx);
}
